;;; -*- Mode: Lisp; Package: mellowcard -*-
;;;
;;; Lambda prolog specials for mellowcard
;;;
;;; Author: Timothy Freeman.  Last Modified Fri Oct  6 04:09:50 1989
;;;
;;; Sccs Id "@(#)mellowlam.lisp	1.12 10/6/89
;;;
;;; ******************************************************************* ;;;
;;;          (c) Copyright 1989 by Carnegie Mellon University.          ;;;
;;;                        All rights reserved.                         ;;;
;;;         This code was written as part of the Ergo project.          ;;;
;;;  If you want to use this code or any Ergo software, please contact  ;;;
;;;			Frank Pfenning (fp@cs.cmu.edu)			;;;
;;; ******************************************************************* ;;;

(in-package :mellowcard)
(use-package :ergolisp)
(use-package :elp-hacks)

;;; This currently has the type o -> string -> o, where the string has
;;; to be a variable.  If the user puts a constant there, we get an error.
;;; If this had the type o -> (string -> o) -> o then the typechecker
;;; would be able to keep the user from making this mistake.  I don't
;;; think this is worth fixing.
(elp::defspecial
 writing_to_string (elp::rgform elp::rterm) t
 ( (list goal stringvar) con ctag sc )
 "Solves goal, then unifies the output generated while this happens with
stringvar.  If we fail back to here, we don't bother trying goal again
because I (TSF) don't know what value to use for stringvar the second time
around. "
 (let* ((old-elp-output elp::*elp-standard-output*)
	(elp::*elp-standard-output* (make-string-output-stream)))
   (catch 'punting-writing-retry
     (elp::solve goal
		 (elp::tlang-value empty-constraint)
		 ctag
		 (elp::sc-make
		  (fdlambda (_)
		    (check-type stringvar lterm::evar)
		    (dlet* ( ((lterm::evar _ type _ _ _) stringvar)
			     (newterm (lterm::from-slterm
				       (mk-str
					(get-output-stream-string
					 elp::*elp-standard-output*))
				       type
				       (elp:currently-accessible-modules)))
			     (elp::*elp-standard-output*
			      old-elp-output) )
		      (elp:tlang-call unify stringvar newterm con sc))
		    (throw 'punting-writing-retry nil)))))))

#|Test case.
use "mellowlow".
writing_to_string (write 3, writesans ' is a number') X.
% Should unify X with "3 is a number".
|#

;; Need to pick a bigger font, and add more entertaining flavors.
;; Also ought to fetch this from the .Xdefaults file.
(defparameter *lprolog-font-table*
  '((:normal . "vtsingle")
    (:bold . "vtbold")
    ;(:symbol . "/usr/ergo/font/ergo-kilter.onx")
    )
  "Alist mapping from mellowcard font keywords to font names.")

(defvar *mouse-message-window*
  nil
  "Window to display mouse messages on.")

(defvar *disp-table* nil
  "The value of display::id-to-handler when we last created
*mouse-message-window*.  This allows us to recover gracefully from
pithings and closed displays.")

;;; Here we evaluate an lprolog goal entirely for side effects.
;;; Disgusting.  I (TSF) am a real common lisp programmer; I can write 
;;; common lisp code in any language, even lambda prolog.
(defun do-default-lprolog-action (default-action text)
  (dlet ( ((values goal _)
	   (elp::tlang-call
	    convert-squery
	    (mk-appl default-action
		     (parse-lprolog-from-string text)))) )
    (catch 'fail-totally
     (elp::solve goal
		 (elp::tlang-value empty-constraint)
		 'fail-totally
		 (fdlambda (_)
		     (throw 'fail-totally 'junk))))))

(defun do-lprolog-action (action)
  (catch 'fail-totally
    (elp::solve
     (elp::tlang-call convert-squery
		      (parse-lprolog-from-string action))
     (elp::tlang-value empty-constraint)
     'fail-totally
     (fdlambda (_)
       (throw 'fail-totally 'junk))))
  (values))

(def-combined-class mellow-mouse-window
  display:vertical-stacking-message-window
  display:mouse-message-mixin
  (anonymous-class mouse-deletion ()
    :init-methods
    ((display:message-window
       client-self)
     (display:message-string "M,R: Delete this window.")
     (display:name "Mouse Help Window"))
    :called-client-methods (display:destroy-window)
    :user-methods
    ((display:key-or-button (key downp code state part)
       (declare (ignore state part))
       (when (and downp (eq key :button) (or (= code display:*right-button*)
					     (= code display:*middle-button*)))
	 (display:destroy-window))))))

(elp::defspecial
 show_mellowcard_text_low
 (elp::rterm elp::rgabst elp::rterm elp::rterm elp::rterm) t
 ( (list title
	 default-action
	 (syntactic-to-logical (mk-str before-message-text))
	 (syntactic-to-logical (mk-str after-message-text))
	 (syntactic-to-logical (mk-str text))) con _ sc )
 "Makes a mellowcard window.  Title is the title, default-action is
what to do if there is no |action keyword, message-generator generates
messages if there is no |message keyword, and text is the text to
display.  Other keywords allowed are |lisp-action (gives a lisp
expression to evaluate), |plaintext (gives text to use when
evaulating ordinarify_mellowcard_text), |font (gives a font keyword),
|nohigh (makes the glyph unhighlightable)."
 (let ((parsed
	(parse-glyphs text
		      :fill-glyph
		      #'(lambda (glyph text special-p
				       &key
				       (key "" key-sp)
				       (action "" action-sp)
				       (message "" message-sp)
				       (lisp-action "" lisp-action-sp)
				       (plaintext "" plaintext-sp)
				       (font "bold")
				       (nohigh "" nohigh-sp))
			  (declare (ignore nohigh))
			  (unless key-sp (setq key text))
			  (setf (glyph-text glyph) text)
			  (setf (glyph-plaintext glyph)
				(if plaintext-sp plaintext text))
			  (setf (glyph-fontkey glyph)
				(if special-p
				    (intern (string-upcase font) "KEYWORD")
				    :normal))
			  (setf (glyph-message glyph)
				(cond
				 (message-sp message)
				 ((not special-p) "")
				 (t (concatenate 'string before-message-text
						 text
						 after-message-text))))
			  (setf (glyph-key glyph)
				(cond
				 (lisp-action-sp
				  (let ((*package* (find-package "USER")))
				    (with-input-from-string (s lisp-action)
				      (read s))))
				 (action-sp
				  `(do-lprolog-action ',action))
				 (t `(do-default-lprolog-action
				      ',(elp:tlang-call
					   gform-to-sterm default-action
					   () () () 0)
				      ',key))))
			  (setf (glyph-pointable glyph)
				(if nohigh-sp nil special-p))))))
   ;; This makes sure that if all windows have been pithed
   ;; since the mouse-window was created, we recreate the mouse-window.
   (display::reopen-display)
   (unless (and *mouse-message-window* (eq *disp-table* display::id-to-handler)
		(not (display:deleted-p *mouse-message-window*)))
     (setq *mouse-message-window*
	   (make-instance mellow-mouse-window)
	   *disp-table* display::id-to-handler))
   (make-instance hypertext-win
		  :title (trim-quotes (format nil "~a" title))
		  :message-window *mouse-message-window*
		  :display-term
		  (make-instance hypertext-dt
				 :glyphs parsed
				 :font-table *lprolog-font-table*)
		  ;; How do we know when we should delete the window?
		  :interpret-fn
		  #'eval))
 (elp::sc-invoke sc con))

#|Test case.
use "mellowlow".
show_mellowcard_text_low "hi" (Thing \ (write Thing)) "Before" "After" "{Text|key \'text\'}".
|#

(defun trim-quotes (string)
  (if (and (>= (length string) 2)
	   (char-equal (elt string 0) #\")
	   (char-equal (elt string (- (length string) 1)) #\"))
      (subseq string 1 (- (length string) 1))
      string))

(elp::defspecial
 ordinarify_mellowcard_text (elp::rterm elp::rterm) t
 ( (list (syntactic-to-logical (mk-str stringin)) stringout) con _ sc )
 "Remove MellowCard formatting punctuation from the given string."
 (dlet* ((parsed (parse-glyphs stringin
			      :fill-glyph
			      #'(lambda (glyph text special-p
					       &key (plaintext ""
							       plaintext-sp)
					       &allow-other-keys)
				  (declare (ignore special-p))
				  (setf (glyph-plaintext glyph)
					(if plaintext-sp plaintext text)))))
	 (ordinary (de-parse-glyphs parsed))
	 ((lterm::evar _ type _ _ _) stringout)
	 (newterm (lterm::from-slterm
		   (mk-str ordinary) type (elp:currently-accessible-modules))))
   (elp:tlang-call unify newterm stringout con sc)))

#|Test case.
use "mellowlow".
ordinarify_mellowcard_text "foo {bar} {baz|plaintext 3}" X.
% Should unify X with "foo bar 3".
|#

(elp::defspecial
 mellowfy_ordinary_text (elp::rterm elp::rterm) t
 ( (list (syntactic-to-logical (mk-str stringin)) stringout) con _ sc )
 "Add MellowCard escapes to the given string."
 (dlet* ((mellowfied (add-mellowcard-escapes stringin))
	 ((lterm::evar _ type _ _ _) stringout)
	 (newterm (lterm::from-slterm
		   (mk-str mellowfied)
		   type (elp:currently-accessible-modules))))
   (elp:tlang-call unify newterm stringout con sc)))

#|Test case.
use "mellowlow".
mellowfy_ordinary_text "a \ b {" X.
% Should unify X with "a \\ b \{".
|#

(in-package :sb-runtime)
;;; This routine is the same as the system routine read-char, except it works on
;;; placestreams and some useless arguments (in this context) are omitted.
(defun place-read-char (self eof-value &aux linetext eof-at-eolp)
  (if (placestream-unreadp self)
      (setf (placestream-unreadp self) nil)
      (setf (placestream-lastch self)
	    (cond ((and (= (placestream-charnumber self)
			   (placestream-linelength self))
			(placestream-eof-at-eolp self))
		   :eof)
		  ((= (placestream-charnumber self)
		      (placestream-linelength self))
		   (setf (placestream-lastline self) (placestream-linetext self))
		   (multiple-value-setq (linetext eof-at-eolp)
		     (let ((stream (placestream-stream self)))
		       (if (and
			    #+lucid
			    (not (typep stream 'lucid::string-input-stream))
			    (equal
			     (display::extract-stream-read-handle stream)
			     (display::extract-stream-read-handle
			      *terminal-io*))
			    (display:display-opened))
			   (display:read-line-mouse-poll stream nil :eof nil)
			   (read-line stream nil :eof nil))))
		   (when (equal linetext :eof)
		     (setq linetext ""
			   eof-at-eolp t))
		   (setf (placestream-linetext self) linetext
			 (placestream-eof-at-eolp self) eof-at-eolp
			 (placestream-linelength self) (length linetext)
			 (placestream-charnumber self) 0)
		   (incf (placestream-linenumber self))
		   #\newline)
		  (t (prog1 (elt (placestream-linetext self)
				 (placestream-charnumber self))
			    (incf (placestream-charnumber self)))))))
  (if (equal (placestream-lastch self) :eof)
      eof-value
      (placestream-lastch self)))
